# Rodent Arena Tracker - By: Jonathan Krynitsky, 2019
#
# Sept 2019 -- made some adjustments so that threshholds are updated when light levels change. This
# can allow for things like tracking across changes in light cycles. -Julia Pai
#
# Feb 2020 -- added RTC module support, Jonathan Krynitsky and Lex Kravitz
#
###################################################################################################

import sensor
import image
import time
import os
import lcd
import math
import pyb
from pyb import I2C

###################################################################################################
# Define  processing functions
def setThresholds(img, threshLimit, mouseColor):        # Set mouse blob threshold
    hist = img.get_histogram()
    prc = hist.get_percentile(threshLimit)
    if mouseColor == 'black':
        thresholds = (0, prc.value(), 0, 55, 0, 55)
    elif mouseColor == 'white':
        thresholds = (prc.value(), 255, 0, 55, 0, 55)
    return thresholds

def processImg(img, thresholds, max_mouse_size, prev_centroid, fileName, value):
    all_blobs = img.find_blobs([thresholds], invert=False, pixels_threshold=40, area_threshold=10, merge=False)
    largest_blob = 0
    large_x = -1
    large_y = -1
    speed = -1

    for blob in all_blobs:
        if blob.pixels() < max_mouse_size:     #ignore massive blobs
            if blob.pixels() > largest_blob:
                largest_blob = blob.pixels()
                large_x = blob.cx()
                large_y = blob.cy()

    if largest_blob > 0:
        speed = int(math.pow(math.pow(large_x - prev_centroid[0], 2) + math.pow(large_y - prev_centroid[1], 2), .5))
        img.draw_cross(large_x, large_y, 50, color=255)
        string_out = 'X:' + str(large_x) + ' Y:' + str(large_y) + ' S:' + str(speed)
        img.draw_string(1, 148, string_out, (255,0,0))

    img.draw_string(1, 139, 'TTL IN:' + str(value), (255,0,0))
    img.draw_string(1, 1 , datetime, (255,0,0))
    img.draw_string(1, 10 ,fileName, (255,0,0))
    lcd.display(img)
    centroid = (large_x, large_y, speed)
    return centroid

###################################################################################################

# Set up Sensor
sensor.reset()
sensor.set_pixformat(sensor.GRAYSCALE)
sensor.set_framesize(sensor.LCD)

# Set up NIR LED
ir_leds = pyb.LED(4)
ir_leds.on()

# Init LCD
lcd.init()

# Init I2C and buffer for RTC
i2c = I2C(2)                         # create on bus 2
i2c = I2C(2, I2C.MASTER)             # create and init as a master
print(i2c)
i2c.init(I2C.MASTER, baudrate=400000) # init as a master

i2c_addr = i2c.scan()
RTC_addr = i2c_addr[0]
RTC_time = bytearray(7)  # create a buffer

# Read RTC time
RTC_time = i2c.mem_read(7, RTC_addr, 0)
seconds = hex(RTC_time[0])[2:]
minutes = hex(RTC_time[1])[2:]
hours = hex(RTC_time[2])[2:]
dayofweek = hex(RTC_time[3])[2:]
dayofmonth = hex(RTC_time[4])[2:]
month = hex(RTC_time[5])[2:]
year = hex(RTC_time[6])[2:]
datetime = (month + "/" + dayofmonth + "/" + year + " " + hours + ":" + minutes + ":" + seconds)

# Prepare File IO - Directory
try:
    os.mkdir('CentroidData')
except OSError as e:
    print('folder already exisits')

# Prepare File IO - Find next output file number
# This function is much cruder than it needs to be due to lack of exist() fcn!?!?!?!
fileNum = 0
while(True):
    try:
        f = open('/CentroidData/data_out_' + str(fileNum) + '.csv')
        f.close()
        fileNum = fileNum + 1
    except OSError as e:
        break
fileName = 'data_out_' + str(fileNum) + '.csv'
f = open('/CentroidData/' + fileName , mode='w')
f.write('Data file for experiment number ' + str(fileNum) + '\r\n')
f.write('Experiment start: ' + month + '/' + dayofmonth + '/' + year + '  ' + hours + ':' + minutes + ':' + seconds + '\r\n')
f.write('DateTime, XPos, YPos, Speed(pixel/sec), TTL in' + '\r\n')

# Set up TTL Input
adc = pyb.ADC(pyb.Pin('P6'))

# Init Variables
prev_centroid = (-1 , -1, -1)
proc_prev_time = pyb.millis()
cap_prev_time = pyb.millis()
frameCount_rethresh = 0
mouseColor = 'black'     # specify as string 'black' or 'white'
max_mouse_size = 20000

# Set up a clock to track FPS
clock = time.clock()

# Set mouse blob thresholds
img = sensor.snapshot()
imStatsPrev = img.get_statistics()
if mouseColor == 'black':
    threshLimit = 0.04
elif mouseColor == 'white':
    threshLimit = 0.96
thresholds =  setThresholds(img, threshLimit, mouseColor)
print(imStatsPrev)
print(thresholds)

# set a limit on framerate in order to stabilize
period = 70;

# initialize millitemp
millitemp = pyb.millis()

# Main Loop
while(True):
    frameCount_rethresh += 1
    clock.tick()

    # Stabalize fps
    cap_curr_time = pyb.millis()
    cap_diff_time = cap_curr_time - cap_prev_time
    if cap_diff_time < period:
       time.sleep(period-cap_diff_time)
    cap_prev_time = pyb.millis()
    last_seconds = seconds

    # every ~8 seconds, check if image statistics have changed and rethreshold
    if frameCount_rethresh % 100 == 0:
        imStats = img.get_statistics()
        print(thresholds)
        if imStats.mean() > (imStatsPrev.mean()+ imStatsPrev.stdev()) or imStats.mean() < (imStatsPrev.mean()-imStatsPrev.stdev()):
            thresholds = setThresholds(img, threshLimit, mouseColor)
            imStatsPrev = imStats
            print('re-thresholded!')
            print(imStatsPrev)
            img.draw_string(1, 79, 'Rethresholding!', (255,255,0))
            lcd.display(img)

    # Read TTL P6 value
    analogIn = adc.read() > 3500

    # Capture frame
    img = sensor.snapshot()

    # Process Frame
    centroid = processImg(img, thresholds, max_mouse_size, prev_centroid, fileName, analogIn)
    prev_centroid = centroid

    # Process timing
    proc_curr_time = pyb.millis()
    proc_diff_time = proc_curr_time - proc_prev_time
    proc_prev_time = proc_curr_time

    #get RTC time
    RTC_time = i2c.mem_read(7, RTC_addr, 0)
    seconds = hex(RTC_time[0])[2:]
    minutes = hex(RTC_time[1])[2:]
    hours = hex(RTC_time[2])[2:]
    dayofweek = hex(RTC_time[3])[2:]
    dayofmonth = hex(RTC_time[4])[2:]
    month = hex(RTC_time[5])[2:]
    year = hex(RTC_time[6])[2:]

    #calculate milliseconds
    if last_seconds != seconds:
        millitemp = pyb.millis()
    milliseconds = pyb.millis() - millitemp
    datetime = (month + "/" + dayofmonth + "/" + year + " " + hours + ":" + minutes + ":" + seconds)
    datetime_w_millis = (month + "/" + dayofmonth + "/" + year + " " + hours + ":" + minutes + ":" + seconds + "." + str(milliseconds))
    print (datetime_w_millis)

    # Output Data
    speed = int(centroid[2] * 1000 / proc_diff_time)    #normalize speed
    f.write(str(datetime_w_millis) + ',' + str(centroid[0]) + ',' + str(centroid[1]) + ',' + str(speed) + ',' + str(analogIn) + '\r\n')
    f.flush()
